from .base import BaseImageProfiler
from .image_profiler import ImageReport


class TifProfiler(BaseImageProfiler):
    def __init__(self):
        super().__init__()

    def probe(self, path) -> ImageReport:
        # TODO: Implement TIF-specific analysis
        return ImageReport(path=str(path), status="not implemented")



import os
import tempfile
from pathlib import Path

import pytest

from filoma.directories import DirectoryProfiler

# Skip these filesystem-heavy tests on CI where external tools like `fd` may be missing
CI = os.getenv("CI") == "true" or os.getenv("GITHUB_ACTIONS") == "true"
pytestmark = pytest.mark.skipif(CI, reason="Skip on CI where fd may be unavailable")


def test_rust_python_consistency():
    """Test that Rust and Python implementations produce consistent results."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        # Create test structure
        # Root (depth 0)
        # ├── level1/ (depth 1)
        # │   ├── file1.txt (depth 2)
        # │   └── level2/ (depth 2)
        # │       ├── file2.txt (depth 3)
        # │       └── level3/ (depth 3)
        # │           └── file3.txt (depth 4)
        (tmp_path / "level1" / "level2" / "level3").mkdir(parents=True)
        (tmp_path / "level1" / "file1.txt").write_text("test")
        (tmp_path / "level1" / "level2" / "file2.txt").write_text("test")
        (tmp_path / "level1" / "level2" / "level3" / "file3.txt").write_text("test")

        # Test both implementations
        python_profiler = DirectoryProfiler(use_rust=False)
        rust_profiler = DirectoryProfiler(use_rust=True)

        # Test without max_depth - should find all files and folders
        result_py = python_profiler.probe(str(tmp_path))
        result_rust = rust_profiler.probe(str(tmp_path))

        # Expected: 3 files (file1.txt, file2.txt, file3.txt)
        # Expected: 4 folders (root, level1, level2, level3)
        assert result_py["summary"]["total_files"] == 3
        assert result_rust["summary"]["total_files"] == 3
        assert result_py["summary"]["total_folders"] == 4
        assert result_rust["summary"]["total_folders"] == 4

        assert result_py["summary"]["total_files"] == result_rust["summary"]["total_files"]
        assert result_py["summary"]["total_folders"] == result_rust["summary"]["total_folders"]

        # Test with max_depth=2
        # Rule: directories at depth <= 2, files at depth <= 3
        # Expected directories: root(0), level1(1), level2(2) = 3 folders
        # Expected files: file1.txt(2), file2.txt(3) = 2 files
        # Excluded: level3(3) directory, file3.txt(4) file
        result_py_depth = python_profiler.probe(str(tmp_path), max_depth=2)
        result_rust_depth = rust_profiler.probe(str(tmp_path), max_depth=2)

        assert result_py_depth["summary"]["total_files"] == 2
        assert result_rust_depth["summary"]["total_files"] == 2
        assert result_py_depth["summary"]["total_folders"] == 3
        assert result_rust_depth["summary"]["total_folders"] == 3
        assert result_py_depth["summary"]["max_depth"] == 2
        assert result_rust_depth["summary"]["max_depth"] == 2

        assert result_py_depth["summary"]["total_files"] == result_rust_depth["summary"]["total_files"]
        assert result_py_depth["summary"]["max_depth"] == result_rust_depth["summary"]["max_depth"]

        # Test with max_depth=1
        # Rule: directories at depth <= 1, files at depth <= 2
        # Expected directories: root(0), level1(1) = 2 folders
        # Expected files: file1.txt(2) = 1 file
        # Excluded: level2(2), level3(3) directories, file2.txt(3), file3.txt(4) files
        result_py_depth1 = python_profiler.probe(str(tmp_path), max_depth=1)
        result_rust_depth1 = rust_profiler.probe(str(tmp_path), max_depth=1)

        assert result_py_depth1["summary"]["total_files"] == 1
        assert result_rust_depth1["summary"]["total_files"] == 1
        assert result_py_depth1["summary"]["total_folders"] == 2
        assert result_rust_depth1["summary"]["total_folders"] == 2
        assert result_py_depth1["summary"]["max_depth"] == 1
        assert result_rust_depth1["summary"]["max_depth"] == 1

        assert result_py_depth1["summary"]["total_files"] == result_rust_depth1["summary"]["total_files"]
        assert result_py_depth1["summary"]["total_folders"] == result_rust_depth1["summary"]["total_folders"]


def test_empty_directory_consistency():
    """Test consistency with empty directories."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        # Create structure with empty directories
        (tmp_path / "empty1").mkdir()
        (tmp_path / "non_empty").mkdir()
        (tmp_path / "non_empty" / "file.txt").write_text("test")
        (tmp_path / "non_empty" / "empty2").mkdir()

        python_profiler = DirectoryProfiler(use_rust=False)
        rust_profiler = DirectoryProfiler(use_rust=True)

        result_py = python_profiler.probe(str(tmp_path))
        result_rust = rust_profiler.probe(str(tmp_path))

        # Should have same counts
        assert result_py["summary"]["total_files"] == result_rust["summary"]["total_files"] == 1
        assert result_py["summary"]["total_folders"] == result_rust["summary"]["total_folders"] == 4  # root, empty1, non_empty, empty2


def test_single_file_consistency():
    """Test consistency with just a single file."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        (tmp_path / "single_file.txt").write_text("test")

        python_profiler = DirectoryProfiler(use_rust=False)
        rust_profiler = DirectoryProfiler(use_rust=True)

        result_py = python_profiler.probe(str(tmp_path))
        result_rust = rust_profiler.probe(str(tmp_path))

        assert result_py["summary"]["total_files"] == result_rust["summary"]["total_files"] == 1
        assert result_py["summary"]["total_folders"] == result_rust["summary"]["total_folders"] == 1  # just root



#!/usr/bin/env python3
"""
Test script for fd integration in filoma.

This script tests the new fd integration capabilities.
"""

import importlib.util
import sys
import warnings
from pathlib import Path

import pytest

# Add src to path so we can import filoma
sys.path.insert(0, str(Path(__file__).parent.parent / "src"))

from filoma.core import FdIntegration
from filoma.directories import DirectoryProfiler, FdFinder

# Detect Rust availability by checking module spec (avoids importing at test collection)
RUST_AVAILABLE_LOCAL = importlib.util.find_spec("filoma.filoma_core") is not None

# Warning regex used by multiple tests
WARNING_REGEX = r"^\[filoma\] Progress bar updates are limited"


def test_fd_integration():
    """Test fd integration components."""
    # Suppress known progress warning emitted by the Rust backend during scans,
    # but only for this test so other tests can still assert it.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=WARNING_REGEX)

        print("🔍 Testing fd Integration\n")

        # Test 1: Check fd availability
        print("1. Checking fd availability...")
        fd = FdIntegration()
    if not fd.is_available():
        pytest.skip("fd not available on this system")
    else:
        print(f"   ✅ fd is available: {fd.get_version()}")

    # Test 2: Basic fd search
    print("\n2. Testing basic fd search...")
    try:
        # Use glob mode for the *.py pattern
        files = fd.find(pattern="*.py", path=".", max_results=5, use_glob=True)
        print(f"   ✅ Found {len(files)} Python files: {files[:3]}{'...' if len(files) > 3 else ''}")
        assert isinstance(files, list)
        assert len(files) > 0
    except Exception as e:
        pytest.fail(f"fd search failed: {e}")

    # Test 3: FdFinder interface
    print("\n3. Testing FdFinder interface...")
    try:
        searcher = FdFinder()
        if searcher.is_available():
            python_files = searcher.find_by_extension(".py", max_depth=2)
            print(f"   ✅ FdFinder found {len(python_files)} Python files")
            assert isinstance(python_files, list)
        else:
            pytest.skip("FdFinder (fd) not available")
    except Exception as e:
        pytest.fail(f"FdFinder failed: {e}")

    # Test 4: DirectoryProfiler with fd backend
    print("\n4. Testing DirectoryProfiler with fd backend...")
    try:
        profiler = DirectoryProfiler(search_backend="fd")
        if not profiler.is_fd_available():
            pytest.skip("DirectoryProfiler fd backend not available")

        print("   ✅ DirectoryProfiler fd backend available")

        # Quick test on current directory
        result = profiler.probe(".", max_depth=2)
        print(f"   ✅ Analysis completed: {result['summary']['total_files']} files found")
        print(f"   ✅ Backend used: {result['timing']['implementation']}")
        assert "summary" in result and "total_files" in result["summary"]
        assert result["summary"]["total_files"] >= 0
        assert "timing" in result and "implementation" in result["timing"]
    except Exception as e:
        pytest.fail(f"DirectoryProfiler fd backend failed: {e}")


def test_backend_comparison():
    #!/usr/bin/env python3
    """
    Test script for fd integration in filoma.

    This script tests the fd integration capabilities.
    """

    import importlib.util
    import sys
    import warnings
    from pathlib import Path

    import pytest

    # Add src to path so we can import filoma
    sys.path.insert(0, str(Path(__file__).parent.parent / "src"))

    from filoma.core import FdIntegration
    from filoma.directories import DirectoryProfiler, FdFinder

    # Detect Rust availability by checking module spec (avoids importing at test collection)
    RUST_AVAILABLE_LOCAL = importlib.util.find_spec("filoma.filoma_core") is not None

    # Warning regex used by multiple tests
    WARNING_REGEX = r"^\[filoma\] Progress bar updates are limited"

    def test_fd_integration():
        """Test fd integration components."""
        # Suppress known progress warning emitted by the Rust backend during scans,
        # but only for this test so other tests can still assert it.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=WARNING_REGEX)

            print("🔍 Testing fd Integration\n")

            # Test 1: Check fd availability
            print("1. Checking fd availability...")
            fd = FdIntegration()

            if not fd.is_available():
                pytest.skip("fd not available on this system")
            else:
                print(f"   ✅ fd is available: {fd.get_version()}")

            # Test 2: Basic fd search
            print("\n2. Testing basic fd search...")
            try:
                # Use glob mode for the *.py pattern
                files = fd.find(pattern="*.py", path=".", max_results=5, use_glob=True)
                print(f"   ✅ Found {len(files)} Python files: {files[:3]}{'...' if len(files) > 3 else ''}")
                assert isinstance(files, list)
                assert len(files) > 0
            except Exception as e:
                pytest.fail(f"fd search failed: {e}")

            # Test 3: FdFinder interface
            print("\n3. Testing FdFinder interface...")
            try:
                searcher = FdFinder()
                if searcher.is_available():
                    python_files = searcher.find_by_extension(".py", max_depth=2)
                    print(f"   ✅ FdFinder found {len(python_files)} Python files")
                    assert isinstance(python_files, list)
                else:
                    pytest.skip("FdFinder (fd) not available")
            except Exception as e:
                pytest.fail(f"FdFinder failed: {e}")

            # Test 4: DirectoryProfiler with fd backend
            print("\n4. Testing DirectoryProfiler with fd backend...")
            try:
                profiler = DirectoryProfiler(search_backend="fd")
                if not profiler.is_fd_available():
                    pytest.skip("DirectoryProfiler fd backend not available")

                print("   ✅ DirectoryProfiler fd backend available")

                # Quick test on current directory
                result = profiler.probe(".", max_depth=2)
                print(f"   ✅ Analysis completed: {result['summary']['total_files']} files found")
                print(f"   ✅ Backend used: {result['timing']['implementation']}")
                assert "summary" in result and "total_files" in result["summary"]
                assert result["summary"]["total_files"] >= 0
                assert "timing" in result and "implementation" in result["timing"]
            except Exception as e:
                pytest.fail(f"DirectoryProfiler fd backend failed: {e}")

    def test_backend_comparison():
        """Compare different backends."""
        print("\n🔬 Backend Performance Comparison\n")

        import time

        test_dir = "."
        max_depth = 2

        backends = []

        # Test available backends
        for backend in ["python", "rust", "fd"]:
            try:
                profiler = DirectoryProfiler(search_backend=backend)

                # Check if backend is actually available
                if backend == "fd" and not profiler.is_fd_available():
                    continue
                elif backend == "rust" and not profiler.is_rust_available():
                    continue

                print(f"Testing {backend} backend...")
                start_time = time.time()
                if backend == "rust":
                    # Suppress the progress warning here so the dedicated warning test can assert it
                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore", message=WARNING_REGEX)
                        result = profiler.probe(test_dir, max_depth=max_depth)
                else:
                    result = profiler.probe(test_dir, max_depth=max_depth)
                elapsed = time.time() - start_time

                backends.append(
                    {
                        "name": backend,
                        "time": elapsed,
                        "files": result["summary"]["total_files"],
                        "folders": result["summary"]["total_folders"],
                    }
                )

                print(f"   {backend}: {elapsed:.3f}s, {result['summary']['total_files']} files")

            except Exception as e:
                print(f"   {backend}: Failed - {e}")

        # Show comparison
        if backends:
            print("\n📊 Results Summary:")
            fastest = min(backends, key=lambda x: x["time"])
            for backend in backends:
                speedup = fastest["time"] / backend["time"]
                if backend == fastest:
                    print(f"   🏆 {backend['name']}: {backend['time']:.3f}s (fastest)")
                else:
                    print(f"   📈 {backend['name']}: {backend['time']:.3f}s ({speedup:.1f}x slower)")

    def test_fd_finder_features():
        """Test FdFinder advanced features."""

        searcher = FdFinder()
        if not searcher.is_available():
            print("❌ fd not available for advanced testing")
            return

        # Test different search types
        tests = [
            ("Python files", lambda: searcher.find_by_extension(".py")),
            ("Recent files (1d)", lambda: searcher.find_recent_files(changed_within="1d")),
            ("Large files (>1k)", lambda: searcher.find_large_files(min_size="1k")),
            ("Empty directories", lambda: searcher.find_empty_directories()),
            ("File count", lambda: searcher.count_files()),
        ]

        for test_name, test_func in tests:
            try:
                result = test_func()
                if isinstance(result, int):
                    print(f"✅ {test_name}: {result}")
                else:
                    print(f"✅ {test_name}: {len(result)} results")
            except Exception as e:
                print(f"❌ {test_name}: Failed - {e}")

    def test_rust_progress_warning_expected():
        """Ensure the Rust progress warning is emitted (if Rust backend is used)."""
        if not RUST_AVAILABLE_LOCAL:
            pytest.skip("Rust backend not available")

        # The profiler emits an informational UserWarning about limited progress updates.
        with pytest.warns(UserWarning, match=WARNING_REGEX):
            profiler = DirectoryProfiler(search_backend="rust")
            _ = profiler.probe(".", max_depth=2)

    def test_rust_progress_warning_suppressed():
        """Run the same profiler probe while explicitly suppressing the progress warning."""
        if not RUST_AVAILABLE_LOCAL:
            pytest.skip("Rust backend not available")

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=WARNING_REGEX)
            profiler = DirectoryProfiler(search_backend="rust")
            _ = profiler.probe(".", max_depth=2)

    if __name__ == "__main__":
        print("🚀 filoma fd Integration Test\n")

        # Run main smoke test
        test_fd_integration()



#!/usr/bin/env python3
"""
Test script to demonstrate the new progress indication and timing features
for the directory profiler.
"""

import os
import tempfile
import time
from pathlib import Path

import pytest

from filoma.directories import DirectoryProfiler

# Skip on CI where external tools like `fd` may be missing
CI = os.getenv("CI") == "true" or os.getenv("GITHUB_ACTIONS") == "true"
pytestmark = pytest.mark.skipif(CI, reason="Skip on CI where fd may be unavailable")

# Use print instead of logger for standalone test script


def test_progress_features():
    """Test the new progress and timing features."""
    print("=" * 80)
    print("Testing Directory Profiler with Progress and Timing")
    print("=" * 80)
    test_dir = Path.cwd()
    configs = [
        {
            "name": "Python Implementation with Progress",
            "use_rust": False,
            "show_progress": True,
        },
        {
            "name": "Rust Sequential with Progress",
            "use_rust": True,
            "use_parallel": False,
            "show_progress": True,
        },
        {
            "name": "Rust Parallel with Progress",
            "use_rust": True,
            "use_parallel": True,
            "show_progress": True,
        },
        {
            "name": "Python Implementation without Progress",
            "use_rust": False,
            "show_progress": False,
        },
    ]
    for config in configs:
        print(f"\n🔍 Testing: {config['name']}")
        print("-" * 50)
        profiler = DirectoryProfiler(
            use_rust=config.get("use_rust", True),
            use_parallel=config.get("use_parallel", True),
            show_progress=config.get("show_progress", True),
            build_dataframe=False,
        )
        impl_info = profiler.get_implementation_info()
        print(f"Implementation: {impl_info}")
        try:
            result = profiler.probe(str(test_dir), max_depth=2)
            profiler.print_summary(result)
            if "timing" in result:
                timing = result["timing"]
                print("⏱️  Timing Details:")
                print(f"   - Elapsed: {timing['elapsed_seconds']:.3f}s")
                print(f"   - Implementation: {timing['implementation']}")
                print(f"   - Speed: {timing['items_per_second']:,.0f} items/sec")
        except Exception as e:
            print(f"❌ Error: {e}")
        print("\n" + "=" * 50)
        time.sleep(0.5)


def test_custom_progress_callback():
    """Test custom progress callback functionality."""
    print("\n🔄 Testing Custom Progress Callback")
    print("-" * 50)

    def custom_callback(message: str, current: int, total: int):
        if total > 0:
            percentage = (current / total) * 100
            print(f"📊 {message} - {current:,}/{total:,} ({percentage:.1f}%)")
        else:
            print(f"📊 {message} - {current:,} items processed")

    profiler = DirectoryProfiler(use_rust=False, show_progress=False, progress_callback=custom_callback)
    test_dir = Path.cwd()
    result = profiler.probe(str(test_dir), max_depth=1)
    print("✅ Custom callback test completed!")
    print(f"   - Found {result['summary']['total_files']} files")
    print(f"   - Found {result['summary']['total_folders']} folders")


def test_large_directory():
    """Test with a larger directory to better see progress indication."""
    print("\n" + "=" * 60)
    print("BONUS TEST: Large Directory Structure")
    print("=" * 60)
    with tempfile.TemporaryDirectory() as temp_dir:
        print(f"Creating test directory structure in {temp_dir}")
        for i in range(5):
            level_dir = Path(temp_dir) / f"level_{i}"
            level_dir.mkdir()
            for j in range(20):
                sub_dir = level_dir / f"subdir_{j}"
                sub_dir.mkdir()
                for k in range(10):
                    file_path = sub_dir / f"file_{k}.txt"
                    file_path.write_text(f"Content of file {k} in {sub_dir}")
        print("Test directory structure created. Starting analysis...")
        profiler = DirectoryProfiler(use_rust=True, use_parallel=True, show_progress=True, build_dataframe=False)
        result = profiler.probe(temp_dir)
        profiler.print_summary(result)


if __name__ == "__main__":
    test_progress_features()
    test_custom_progress_callback()
    # Uncomment to run the large directory test
    # test_large_directory()



from collections.abc import Mapping
from dataclasses import asdict, dataclass
from typing import Optional, Tuple

import numpy as np


@dataclass
class ImageReport(Mapping):
    """Structured container for image analysis results."""

    path: Optional[str] = None
    file_type: Optional[str] = None
    shape: Optional[Tuple[int, ...]] = None
    dtype: Optional[str] = None
    min: Optional[float] = None
    max: Optional[float] = None
    mean: Optional[float] = None
    nans: int = 0
    infs: int = 0
    unique: int = 0
    status: Optional[str] = None

    def to_dict(self) -> dict:
        d = asdict(self)
        # Ensure shape is a plain tuple (JSON-serializable)
        if isinstance(self.shape, (list, tuple)):
            d["shape"] = tuple(self.shape)
        return d

    def as_dict(self) -> dict:
        return self.to_dict()

    # Mapping protocol for dict access
    def _as_dict(self) -> dict:
        return self.to_dict()

    def __getitem__(self, key):
        return self._as_dict()[key]

    def __iter__(self):
        return iter(self._as_dict())

    def __len__(self):
        return len(self._as_dict())


class ImageProfiler:
    """
    Provides common analysis methods for image data loaded as numpy arrays.
    Returns an :class:`ImageReport` dataclass for consistency with the rest of the API.
    """

    def probe(self, arr: np.ndarray) -> ImageReport:
        report = ImageReport(
            shape=tuple(arr.shape) if hasattr(arr, "shape") else None,
            dtype=str(arr.dtype) if hasattr(arr, "dtype") else None,
            min=float(np.nanmin(arr)) if arr.size > 0 else None,
            max=float(np.nanmax(arr)) if arr.size > 0 else None,
            mean=float(np.nanmean(arr)) if arr.size > 0 else None,
            nans=int(np.isnan(arr).sum()),
            infs=int(np.isinf(arr).sum()),
            unique=int(np.unique(arr).size) if arr.size > 0 else 0,
        )
        return report



from PIL import Image

from .base import BaseImageProfiler
from .image_profiler import ImageProfiler, ImageReport


class PngProfiler(BaseImageProfiler):
    def __init__(self):
        super().__init__()

    def probe(self, path) -> ImageReport:
        # Load PNG as numpy array
        img = Image.open(path)
        arr = __import__("numpy").array(img)
        profiler = ImageProfiler()
        report = profiler.probe(arr)
        # set file metadata
        report.file_type = "png"
        report.path = str(path)
        return report



"""
Simple ML-style utilities for filoma DataFrame splitting.

Provides an intuitive auto_split API to split a filoma.DataFrame into train/val/test
based on filename/path-derived features. The goal is a tiny, dependency-free,
user-friendly interface using pathlib.Path to select path parts.
"""

from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Union

import polars as pl
from loguru import logger


def _normalize_ratios(ratios: Sequence[int]) -> List[float]:
    total = sum(ratios)
    if total <= 0:
        raise ValueError("Ratios must sum to a positive value")
    return [r / total for r in ratios]


def _stable_hash(s: str, seed: Optional[int] = None) -> int:
    """Return a stable integer hash for a string. Optionally incorporate a seed."""
    m = hashlib.sha256()
    if seed is not None:
        m.update(str(seed).encode("utf-8"))
    m.update(s.encode("utf-8"))
    # Use 8 bytes to get a large integer
    return int.from_bytes(m.digest()[:8], "big")


def _get_feature_value(path_str: str, how: str, parts: Optional[Iterable[int]] = None) -> str:
    p = Path(path_str)
    if how == "parts":
        # parts is an iterable of part indices (negative allowed)
        if parts is None:
            raise ValueError("parts must be provided when how='parts'")
        parts_list = list(parts)
        selected = []
        for idx in parts_list:
            try:
                selected.append(p.parts[idx])
            except IndexError:
                selected.append("")
        return "/".join(selected)
    elif how == "filename":
        return p.name
    elif how == "stem":
        return p.stem
    elif how == "parent":
        return str(p.parent)
    elif how == "suffix":
        return p.suffix
    else:
        raise ValueError(f"Unknown how='{how}'")


def discover_filename_features(
    pl_df: pl.DataFrame,
    sep: str = "_",
    prefix: Optional[str] = "feat",
    max_tokens: Optional[int] = None,
    include_parent: bool = False,
    include_all_parts: bool = False,
    token_names: Optional[Union[str, Sequence[str]]] = None,
    path_col: str = "path",
) -> pl.DataFrame:
    """
    Discover separator-based tokens from filename stems and add them as columns.

    Args:
        pl_df: Polars DataFrame with a 'path' column.
        sep: separator to split the stem (default '_').
        prefix: prefix for generated feature columns (columns will be prefix1, prefix2, ...).
        max_tokens: optional max number of tokens to extract (otherwise uses observed max).
        include_parent: if True, add a 'parent' column containing the immediate parent folder name.

    Returns:
        Polars DataFrame with added feature columns.
    """
    if path_col not in pl_df.columns:
        raise ValueError(f"DataFrame must have a '{path_col}' column")

    # Extract stems and split by sep to discover token counts
    stems = [Path(s).stem for s in pl_df[path_col].to_list()]
    split_tokens = [stem.split(sep) if stem is not None else [""] for stem in stems]
    observed_max = max((len(t) for t in split_tokens), default=0)
    if max_tokens is None:
        max_tokens = observed_max

    # Normalize token_names
    # token_names may be: None, 'auto', or a sequence of names
    if token_names == "auto":
        token_names_seq = None
        auto_mode = True
    elif isinstance(token_names, (list, tuple)):
        token_names_seq = list(token_names)
        auto_mode = False
    else:
        token_names_seq = None
        auto_mode = False

    # For each token index create a column
    new_cols = []
    for i in range(max_tokens):
        # Decide column name: explicit token_names > prefix > generic token
        if token_names_seq is not None and i < len(token_names_seq) and token_names_seq[i]:
            col_name = token_names_seq[i]
        elif auto_mode:
            # auto generates readable token names using 'token' base or prefix if present
            base = prefix if prefix else "token"
            col_name = f"{base}{i + 1}"
        else:
            if prefix:
                col_name = f"{prefix}{i + 1}"
            else:
                col_name = f"token{i + 1}"

        def pick_token(s: str, idx=i):
            st = Path(s).stem
            parts = st.split(sep) if st is not None else [""]
            try:
                return parts[idx]
            except Exception:
                return ""

        new_cols.append(pl.col(path_col).map_elements(pick_token, return_dtype=pl.Utf8).alias(col_name))

    if include_parent:
        new_cols.append(pl.col(path_col).map_elements(lambda s: Path(s).parent.name, return_dtype=pl.Utf8).alias("parent"))

    # Optionally add all path parts as features (path_part0 is root/first part)
    if include_all_parts:
        parts_lists = [list(Path(s).parts) for s in pl_df[path_col].to_list()]
        max_parts = max((len(p) for p in parts_lists), default=0)
        for i in range(max_parts):
            col_name = f"path_part{i}"

            def pick_part(s: str, idx=i):
                try:
                    parts = list(Path(s).parts)
                    return parts[idx]
                except Exception:
                    return ""

            new_cols.append(pl.col(path_col).map_elements(pick_part, return_dtype=pl.Utf8).alias(col_name))

    return pl_df.with_columns(new_cols)


# ------------ Internal helper functions for modular auto_split ------------ #
def _maybe_discover(
    pl_df: pl.DataFrame,
    discover: bool,
    sep: str,
    feat_prefix: str,
    max_tokens: Optional[int],
    include_parent: bool,
    include_all_parts: bool,
    token_names: Optional[Union[str, Sequence[str]]],
    path_col: str,
) -> pl.DataFrame:
    if not discover:
        return pl_df
    return discover_filename_features(
        pl_df,
        sep=sep,
        prefix=feat_prefix,
        max_tokens=max_tokens,
        include_parent=include_parent,
        include_all_parts=include_all_parts,
        token_names=token_names,
        path_col=path_col,
    )


def _build_feature_index(pl_df: pl.DataFrame, path_col: str, how: str, parts: Optional[Iterable[int]]) -> Tuple[dict, List[str]]:
    paths = pl_df[path_col].to_list()
    mapping: dict = {}
    for i, p in enumerate(paths):
        feat = _get_feature_value(p, how=how, parts=parts)
        mapping.setdefault(feat, []).append(i)
    return mapping, paths


def _assign_features(feature_to_idxs: dict, ratios: Sequence[float], seed: Optional[int]) -> dict:
    assignment = {}
    r0, r1 = ratios[0], ratios[0] + ratios[1]
    for feat in feature_to_idxs:
        h = _stable_hash(feat, seed=seed)
        frac = (h % (10**8)) / 1e8
        if frac < r0:
            assignment[feat] = "train"
        elif frac < r1:
            assignment[feat] = "val"
        else:
            assignment[feat] = "test"
    return assignment


def _mask_from_assignment(feature_to_idxs: dict, feature_assignment: dict, total: int) -> List[str]:
    mask: List[str] = [None] * total  # type: ignore
    for feat, idxs in feature_to_idxs.items():
        split = feature_assignment[feat]
        for i in idxs:
            mask[i] = split  # type: ignore
    return mask


def _add_feature_column(pl_df: pl.DataFrame, path_col: str, how: str, parts: Optional[Iterable[int]]) -> pl.DataFrame:
    feat_name = "_feat_parts" if how == "parts" else f"_feat_{how}"
    return pl_df.with_columns(
        [
            pl.col(path_col)
            .map_elements(
                lambda x: _get_feature_value(x, how=how, parts=parts),
                return_dtype=pl.Utf8,
            )
            .alias(feat_name)
        ]
    )


def _maybe_log_ratio_drift(
    train_n: int,
    val_n: int,
    test_n: int,
    total: int,
    ratios: Sequence[float],
    verbose: bool,
):
    if not verbose or total == 0:
        return
    req = ratios
    act_counts = (train_n, val_n, test_n)
    act = tuple(c / total for c in act_counts)
    if any(abs(a - r) > max(1 / total, 0.05) for a, r in zip(act, req)):
        req_pct = ",".join(f"{r * 100:.1f}%" for r in req)
        act_pct = ",".join(f"{a * 100:.1f}%" for a in act)
        logger.warning(
            ("filoma.ml.auto_split: requested ratios {} -> achieved counts {} ({}) vs requested ({}) total={} (grouped hashing can cause drift)"),
            req_pct,
            act_counts,
            act_pct,
            req_pct,
            total,
        )


def auto_split(
    df: Union[pl.DataFrame, Any],
    train_val_test: Tuple[int, int, int] = (80, 10, 10),
    how: str = "parts",
    parts: Optional[Iterable[int]] = (-1,),
    seed: Optional[int] = None,
    discover: bool = False,
    sep: str = "_",
    feat_prefix: str = "feat",
    max_tokens: Optional[int] = None,
    include_parent: bool = False,
    include_all_parts: bool = False,
    token_names: Optional[Union[str, Sequence[str]]] = None,
    path_col: str = "path",
    verbose: bool = True,
    return_type: str = "polars",
) -> Tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
    """
    Split a filoma DataFrame into train/val/test based on filename/path-derived features.

    Args:
        df: a Polars DataFrame or filoma.DataFrame wrapper containing a 'path' column.
        train_val_test: three integers or ratios for train/val/test; they will be
                        normalized to fractions.
        how: which feature to derive from the path. Options: 'parts', 'filename',
             'stem', 'parent', 'suffix'. If 'parts', use the `parts` argument.
        parts: an iterable of integers selecting Path.parts indices (supports negative
               indices). Only used when how='parts'. Default picks -1 (filename).
        seed: optional integer to alter hashing for reproducible, different splits.
        discover: if True, automatically discover filename tokens and add columns
            named `prefix1`, `prefix2`, ... (or `token1`... if prefix=None).
        sep: separator used to split filename stems when `discover=True`.
        feat_prefix: prefix to use for discovered token column names. If None,
            discovered token columns will be named `token1`, `token2`, ...
        token_names: optional list of column names to use for tokens, or 'auto' to
                    automatically generate readable names (uses prefix if set).
        max_tokens: maximum number of tokens to extract when discovering.
        include_parent: if True, add a `parent` column with the immediate parent folder name.
        include_all_parts: if True, add columns `path_part0`, `path_part1`, ... for all Path.parts.
        verbose: if True (default) log a short warning when achieved split counts
            differ noticeably from requested ratios (common with small datasets or
            grouped features).
        return_type: one of 'polars' (default), 'filoma' (wrap Polars into filoma.DataFrame),
                or 'pandas' (convert to pandas.DataFrame). If 'pandas' is chosen,
                pandas must be available.

        Returns:
            (train_df, val_df, test_df) as Polars DataFrames.

        Notes:
            - Splits are deterministic and grouped by the chosen feature to avoid
            leaking similar files into multiple sets when they share the same feature.
            - The method uses sha256 hashing of the feature string to map to [0,1).
    """
    # Accept filoma.DataFrame wrapper or raw Polars DataFrame
    if hasattr(df, "df"):
        pl_df = df.df
    else:
        pl_df = df

    if path_col not in pl_df.columns:
        raise ValueError(f"DataFrame must have a '{path_col}' column")

    ratios = _normalize_ratios(train_val_test)

    # Discovery
    pl_work = _maybe_discover(
        pl_df,
        discover=discover,
        sep=sep,
        feat_prefix=feat_prefix,
        max_tokens=max_tokens,
        include_parent=include_parent,
        include_all_parts=include_all_parts,
        token_names=token_names,
        path_col=path_col,
    )

    # Feature grouping & assignment
    feature_to_idxs, paths = _build_feature_index(pl_work, path_col=path_col, how=how, parts=parts)
    feature_assignment = _assign_features(feature_to_idxs, ratios=ratios, seed=seed)
    mask = _mask_from_assignment(feature_to_idxs, feature_assignment, total=len(paths))
    tmp = pl_work.with_columns([pl.Series("_split", mask)])

    # Feature column for user convenience
    tmp = _add_feature_column(tmp, path_col=path_col, how=how, parts=parts)

    # Split
    train_df = tmp.filter(pl.col("_split") == "train").drop("_split")
    val_df = tmp.filter(pl.col("_split") == "val").drop("_split")
    test_df = tmp.filter(pl.col("_split") == "test").drop("_split")

    _maybe_log_ratio_drift(len(train_df), len(val_df), len(test_df), len(paths), ratios, verbose)

    # Return requested type
    if return_type == "polars" or return_type is None:
        return train_df, val_df, test_df

    if return_type == "filoma":
        # Lazy import filoma.DataFrame wrapper to avoid heavy imports at module import time
        try:
            from .dataframe import DataFrame as FDataFrame
        except Exception:
            from filoma.dataframe import DataFrame as FDataFrame

        return FDataFrame(train_df), FDataFrame(val_df), FDataFrame(test_df)

    if return_type == "pandas":
        try:
            return train_df.to_pandas(), val_df.to_pandas(), test_df.to_pandas()
        except Exception as e:
            raise RuntimeError(f"Failed to convert to pandas DataFrame: {e}")

    raise ValueError(f"Unknown return_type='{return_type}'")



"""
DataFrame module for filoma - provides enhanced data manipulation capabilities
for file and directory analysis results using Polars.
"""

import json
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import polars as pl
from loguru import logger

from .files.file_profiler import FileProfiler


class DataFrame:
    """
    A wrapper around Polars DataFrame for enhanced file and directory analysis.

    This class provides a specialized interface for working with file path data,
    allowing for easy manipulation and analysis of filesystem information.

    All standard Polars DataFrame methods and properties are available through
    attribute delegation, so you can use this like a regular Polars DataFrame
    with additional file-specific functionality.
    """

    def __init__(self, data: Optional[Union[pl.DataFrame, List[str], List[Path]]] = None):
        """
        Initialize a DataFrame.

        Args:
            data: Initial data. Can be:
                - A Polars DataFrame with a 'path' column
                - A list of string paths
                - A list of Path objects
                - None for an empty DataFrame
        """
        if data is None:
            # Default empty DataFrame with a path column for filesystem use-cases
            self._df = pl.DataFrame({"path": []}, schema={"path": pl.String})
        elif isinstance(data, pl.DataFrame):
            # Accept any Polars DataFrame schema. Some operations (path helpers)
            # expect a 'path' column, but group/aggregation helpers return
            # summary tables without 'path' and should still be wrapped.
            self._df = data
        elif isinstance(data, list):
            # Convert to string paths
            paths = [str(path) for path in data]
            self._df = pl.DataFrame({"path": paths})
        else:
            raise ValueError("data must be a Polars DataFrame, list of paths, or None")

    def __getattr__(self, name: str) -> Any:
        """
        Delegate attribute access to the underlying Polars DataFrame.

        This allows direct access to all Polars DataFrame methods and properties
        like columns, dtypes, shape, select, filter, group_by, etc.
        """
        attr = getattr(self._df, name)

        # If it's a method that returns a DataFrame, wrap Polars DataFrame
        if callable(attr):

            def wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                # If the result is a Polars DataFrame, wrap it in filoma.DataFrame
                if isinstance(result, pl.DataFrame):
                    return DataFrame(result)
                # Otherwise return the result as-is
                return result

            return wrapper
        else:
            # For properties and non-callable attributes, return as-is
            return attr

    @property
    def df(self) -> pl.DataFrame:
        """Get the underlying Polars DataFrame."""
        return self._df

    def __len__(self) -> int:
        """Get the number of rows in the DataFrame."""
        return len(self._df)

    def __repr__(self) -> str:
        """String representation of the DataFrame."""
        return f"filoma.DataFrame with {len(self)} rows\n{self._df}"

    def __str__(self) -> str:
        """String representation of the DataFrame."""
        return self.__repr__()

    def head(self, n: int = 5) -> pl.DataFrame:
        """Get the first n rows."""
        return self._df.head(n)

    def tail(self, n: int = 5) -> pl.DataFrame:
        """Get the last n rows."""
        return self._df.tail(n)

    def add_path_components(self) -> "DataFrame":
        """
        Add columns for path components (parent, name, stem, suffix).

        Returns:
            New DataFrame with additional path component columns
        """
        df_with_components = self._df.with_columns(
            [
                pl.col("path").map_elements(lambda x: str(Path(x).parent), return_dtype=pl.String).alias("parent"),
                pl.col("path").map_elements(lambda x: Path(x).name, return_dtype=pl.String).alias("name"),
                pl.col("path").map_elements(lambda x: Path(x).stem, return_dtype=pl.String).alias("stem"),
                pl.col("path").map_elements(lambda x: Path(x).suffix, return_dtype=pl.String).alias("suffix"),
            ]
        )
        return DataFrame(df_with_components)

    def add_file_stats_cols(self, path: str = "path", base_path: Optional[Union[str, Path]] = None) -> "DataFrame":
        """
        Add file statistics columns (size, modified time, etc.) based on a column
        containing filesystem paths.

        Args:
            path: Name of the column containing file system paths.
            base_path: Optional base path. If provided, any non-absolute paths in the
                       path column are resolved relative to this base.

        Returns:
            New DataFrame with file statistics columns added.

        Raises:
            ValueError: If the specified path column does not exist.
        """
        if path not in self._df.columns:
            raise ValueError(f"Column '{path}' not found in DataFrame")

        # Resolve base path if provided
        base = Path(base_path) if base_path is not None else None

        # Use filoma's FileProfiler to collect rich file metadata
        profiler = FileProfiler()

        def get_file_stats(path_str: str) -> Dict[str, Any]:
            try:
                p = Path(path_str)
                if base is not None and not p.is_absolute():
                    p = base / p
                full_path = str(p)
                if not p.exists():
                    logger.warning(f"Path does not exist: {full_path}")
                    return {
                        "size_bytes": None,
                        "modified_time": None,
                        "created_time": None,
                        "is_file": None,
                        "is_dir": None,
                        "owner": None,
                        "group": None,
                        "mode_str": None,
                        "inode": None,
                        "nlink": None,
                        "sha256": None,
                        "xattrs": "{}",
                    }

                # Use the profiler; let it handle symlinks and permissions
                filo = profiler.probe(full_path, compute_hash=False)
                row = filo.as_dict()

                # Normalize keys to a stable schema used by this helper
                return {
                    "size_bytes": row.get("size"),
                    "modified_time": row.get("modified"),
                    "created_time": row.get("created"),
                    "is_file": row.get("is_file"),
                    "is_dir": row.get("is_dir"),
                    "owner": row.get("owner"),
                    "group": row.get("group"),
                    "mode_str": row.get("mode_str"),
                    "inode": row.get("inode"),
                    "nlink": row.get("nlink"),
                    "sha256": row.get("sha256"),
                    "xattrs": json.dumps(row.get("xattrs") or {}),
                }
            except Exception:
                # On any error, return a row of Nones/empties preserving schema
                return {
                    "size_bytes": None,
                    "modified_time": None,
                    "created_time": None,
                    "is_file": None,
                    "is_dir": None,
                    "owner": None,
                    "group": None,
                    "mode_str": None,
                    "inode": None,
                    "nlink": None,
                    "sha256": None,
                    "xattrs": "{}",
                }

        stats_data = [get_file_stats(p) for p in self._df[path].to_list()]

        stats_df = pl.DataFrame(
            stats_data,
            schema={
                "size_bytes": pl.Int64,
                "modified_time": pl.String,
                "created_time": pl.String,
                "is_file": pl.Boolean,
                "is_dir": pl.Boolean,
                "owner": pl.String,
                "group": pl.String,
                "mode_str": pl.String,
                "inode": pl.Int64,
                "nlink": pl.Int64,
                "sha256": pl.String,
                "xattrs": pl.String,
            },
        )

        df_with_stats = pl.concat([self._df, stats_df], how="horizontal")
        return DataFrame(df_with_stats)

    def add_depth_col(self, path: Optional[Union[str, Path]] = None) -> "DataFrame":
        """
        Add a depth column showing the nesting level of each path.

        Args:
            path: The path to calculate depth from. If None, uses the common root.

        Returns:
            New DataFrame with depth column
        """
        if path is None:
            # Find the common root path
            paths = [Path(p) for p in self._df["path"].to_list()]
            if not paths:
                path = Path()
            else:
                # Find common parent
                common_parts = []
                first_parts = paths[0].parts
                for i, part in enumerate(first_parts):
                    if all(len(p.parts) > i and p.parts[i] == part for p in paths):
                        common_parts.append(part)
                    else:
                        break
                path = Path(*common_parts) if common_parts else Path()
        else:
            path = Path(path)

        # Use a different local name to avoid shadowing the parameter inside calculate_depth
        path_root = path

        def calculate_depth(path_str: str) -> int:
            """Calculate the depth of a path relative to the provided root path."""
            try:
                p = Path(path_str)
                relative_path = p.relative_to(path_root)
                return len(relative_path.parts)
            except ValueError:
                # Path is not relative to the provided root path
                return len(Path(path_str).parts)

        df_with_depth = self._df.with_columns([pl.col("path").map_elements(calculate_depth, return_dtype=pl.Int64).alias("depth")])
        return DataFrame(df_with_depth)

    def filter_by_extension(self, extensions: Union[str, List[str]]) -> "DataFrame":
        """
        Filter the DataFrame to only include files with specific extensions.

        Args:
            extensions: File extension(s) to filter by (with or without leading dot)

        Returns:
            Filtered DataFrame
        """
        if isinstance(extensions, str):
            extensions = [extensions]

        # Normalize extensions (ensure they start with a dot)
        normalized_extensions = []
        for ext in extensions:
            if not ext.startswith("."):
                ext = "." + ext
            normalized_extensions.append(ext.lower())

        filtered_df = self._df.filter(pl.col("path").map_elements(lambda x: Path(x).suffix.lower() in normalized_extensions, return_dtype=pl.Boolean))
        return DataFrame(filtered_df)

    def filter_by_pattern(self, pattern: str) -> "DataFrame":
        """
        Filter the DataFrame by path pattern.

        Args:
            pattern: Pattern to match (uses Polars string contains)

        Returns:
            Filtered DataFrame
        """
        filtered_df = self._df.filter(pl.col("path").str.contains(pattern))
        return DataFrame(filtered_df)

    def group_by_extension(self) -> pl.DataFrame:
        """
        Group files by extension and count them.

        Returns:
            Polars DataFrame with extension counts
        """
        df_with_ext = self._df.with_columns(
            [
                pl.col("path")
                .map_elements(lambda x: Path(x).suffix.lower() if Path(x).suffix else "<no extension>", return_dtype=pl.String)
                .alias("extension")
            ]
        )
        result = df_with_ext.group_by("extension").len().sort("len", descending=True)
        return DataFrame(result)

    def group_by_directory(self) -> pl.DataFrame:
        """
        Group files by their parent directory and count them.

        Returns:
            Polars DataFrame with directory counts
        """
        df_with_parent = self._df.with_columns(
            [pl.col("path").map_elements(lambda x: str(Path(x).parent), return_dtype=pl.String).alias("parent_dir")]
        )
        result = df_with_parent.group_by("parent_dir").len().sort("len", descending=True)
        return DataFrame(result)

    def to_polars(self) -> pl.DataFrame:
        """Get the underlying Polars DataFrame."""
        return self._df

    def to_dict(self) -> Dict[str, List]:
        """Convert to a dictionary."""
        return self._df.to_dict(as_series=False)

    def save_csv(self, path: Union[str, Path]) -> None:
        """Save the DataFrame to CSV."""
        self._df.write_csv(str(path))

    def save_parquet(self, path: Union[str, Path]) -> None:
        """Save the DataFrame to Parquet format."""
        self._df.write_parquet(str(path))

    # Convenience methods for common Polars operations that users expect
    @property
    def columns(self) -> List[str]:
        """Get column names."""
        return self._df.columns

    @property
    def dtypes(self) -> List[pl.DataType]:
        """Get column data types."""
        return self._df.dtypes

    @property
    def shape(self) -> tuple:
        """Get DataFrame shape (rows, columns)."""
        return self._df.shape

    def describe(self, percentiles: Optional[List[float]] = None) -> pl.DataFrame:
        """
        Generate descriptive statistics.

        Args:
            percentiles: List of percentiles to include (default: [0.25, 0.5, 0.75])
        """
        return self._df.describe(percentiles=percentiles)

    def info(self) -> None:
        """Print concise summary of the DataFrame."""
        print("filoma.DataFrame")
        print(f"Shape: {self.shape}")
        print(f"Columns: {len(self.columns)}")
        print()

        # Column info
        print("Column details:")
        for i, (col, dtype) in enumerate(zip(self.columns, self.dtypes)):
            null_count = self._df[col].null_count()
            print(f"  {i:2d}  {col:15s} {str(dtype):15s} {null_count:8d} nulls")

        # Memory usage approximation
        memory_mb = sum(self._df[col].estimated_size("mb") for col in self.columns)
        print(f"\nEstimated memory usage: {memory_mb:.2f} MB")

    def unique(self, subset: Optional[Union[str, List[str]]] = None) -> "DataFrame":
        """
        Get unique rows.

        Args:
            subset: Column name(s) to consider for uniqueness
        """
        if subset is None:
            result = self._df.unique()
        else:
            result = self._df.unique(subset=subset)
        return DataFrame(result)

    def sort(self, by: Union[str, List[str]], descending: bool = False) -> "DataFrame":
        """
        Sort the DataFrame.

        Args:
            by: Column name(s) to sort by
            descending: Sort in descending order
        """
        result = self._df.sort(by, descending=descending)
        return DataFrame(result)

    # -------------------- ML convenience API -------------------- #
    def auto_split(
        self,
        train_val_test: Tuple[int, int, int] = (80, 10, 10),
        how: str = "parts",
        parts: Optional[Iterable[int]] = (-1,),
        seed: Optional[int] = None,
        discover: bool = False,
        sep: str = "_",
        feat_prefix: str = "feat",
        max_tokens: Optional[int] = None,
        include_parent: bool = False,
        include_all_parts: bool = False,
        token_names: Optional[Union[str, Sequence[str]]] = None,
        path_col: str = "path",
        verbose: bool = True,
        return_type: str = "filoma",
    ):
        """Deterministically split this filoma DataFrame into train/val/test.

        This is a thin wrapper around ``filoma.ml.auto_split`` so you can call
        ``df.auto_split(...)`` directly on a filoma DataFrame instance.

        Args mirror :func:`filoma.ml.auto_split` except ``df`` is implicit.

        By default ``return_type='filoma'`` so the three returned objects are
        filoma.DataFrame wrappers.
        """
        # Local import to avoid loading ml utilities unless used
        from . import ml  # type: ignore

        return ml.auto_split(
            self,
            train_val_test=train_val_test,
            how=how,
            parts=parts,
            seed=seed,
            discover=discover,
            sep=sep,
            feat_prefix=feat_prefix,
            max_tokens=max_tokens,
            include_parent=include_parent,
            include_all_parts=include_all_parts,
            token_names=token_names,
            path_col=path_col,
            verbose=verbose,
            return_type=return_type,
        )



#!/usr/bin/env python3
"""
Comprehensive backend testing for filoma.

This test suite validates all three backends (Python, Rust, fd) across different scenarios
to ensure they work correctly and provide consistent results where expected.
"""

import tempfile
import time
from pathlib import Path

import pytest

from filoma.core import FdIntegration
from filoma.directories import DirectoryProfiler, FdFinder


class TestBackendComprehensive:
    """Comprehensive backend testing suite."""

    @pytest.fixture
    def test_directory(self):
        """Create a complex test directory structure."""
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)

            # Create a complex directory structure
            # Root
            # ├── docs/
            # │   ├── README.md
            # │   ├── guide.txt
            # │   └── images/
            # │       ├── logo.png
            # │       └── diagram.jpg
            # ├── src/
            # │   ├── main.py
            # │   ├── utils.py
            # │   └── modules/
            # │       ├── __init__.py
            # │       ├── core.py
            # │       └── helpers.rs
            # ├── tests/
            # │   ├── test_main.py
            # │   └── test_utils.py
            # ├── empty_dir/
            # ├── data/
            # │   ├── large_file.txt (1KB)
            # │   └── small.json
            # └── .hidden/
            #     └── secret.txt

            # Create directories
            dirs = ["docs", "docs/images", "src", "src/modules", "tests", "empty_dir", "data", ".hidden"]
            for dir_name in dirs:
                (tmp_path / dir_name).mkdir(parents=True, exist_ok=True)

            # Create files with different sizes and types
            files = {
                "docs/README.md": "# Project\nThis is a test project.",
                "docs/guide.txt": "User guide content here.",
                "docs/images/logo.png": "fake png data" * 10,
                "docs/images/diagram.jpg": "fake jpg data" * 20,
                "src/main.py": "def main():\n    print('hello')\n\nif __name__ == '__main__':\n    main()",
                "src/utils.py": "def helper():\n    return True",
                "src/modules/__init__.py": "",
                "src/modules/core.py": "class Core:\n    pass",
                "src/modules/helpers.rs": "fn helper() -> bool { true }",
                "tests/test_main.py": "def test_main():\n    assert True",
                "tests/test_utils.py": "def test_helper():\n    assert True",
                "data/large_file.txt": "x" * 1000,  # 1KB file
                "data/small.json": '{"key": "value"}',
                ".hidden/secret.txt": "secret content",
            }

            for file_path, content in files.items():
                (tmp_path / file_path).write_text(content)

            yield str(tmp_path)

    def test_python_backend_basic(self, test_directory):
        """Test Python backend basic functionality."""
        profiler = DirectoryProfiler(search_backend="python", show_progress=False)
        result = profiler.probe(test_directory)

        # Verify basic counts
        assert result["summary"]["total_files"] >= 12  # At least our created files
        assert result["summary"]["total_folders"] >= 8  # Including root and created dirs
        assert result["summary"]["empty_folder_count"] >= 1  # empty_dir

        # Verify extensions are detected
        extensions = result["file_extensions"]
        assert ".py" in extensions
        assert ".md" in extensions
        assert ".txt" in extensions
        assert ".rs" in extensions

        # Verify empty folders are detected
        empty_folders = [Path(p).name for p in result["empty_folders"]]
        assert "empty_dir" in empty_folders

    @pytest.mark.skipif(not hasattr(DirectoryProfiler(show_progress=False), "_probe_rust"), reason="Rust backend not available")
    def test_rust_backend_basic(self, test_directory):
        """Test Rust backend basic functionality."""
        profiler = DirectoryProfiler(search_backend="rust", show_progress=False)
        result = profiler.probe(test_directory)

        # Verify basic counts
        assert result["summary"]["total_files"] >= 12
        assert result["summary"]["total_folders"] >= 8

        # Verify extensions are detected
        extensions = result["file_extensions"]
        assert ".py" in extensions
        assert ".md" in extensions
        assert ".txt" in extensions

    def test_fd_backend_basic(self, test_directory):
        """Test fd backend basic functionality."""
        fd = FdIntegration()
        if not fd.is_available():
            pytest.skip("fd not available")

        profiler = DirectoryProfiler(search_backend="fd", show_progress=False)
        result = profiler.probe(test_directory)

        # Verify basic counts
        assert result["summary"]["total_files"] >= 12
        assert result["summary"]["total_folders"] >= 8

    def test_backend_consistency(self, test_directory):
        """Test that different backends produce consistent results."""
        fd = FdIntegration()

        # Get results from available backends
        results = {}

        # Python (always available)
        profiler_py = DirectoryProfiler(search_backend="python", show_progress=False)
        results["python"] = profiler_py.probe(test_directory)

        # Rust (if available)
        try:
            profiler_rust = DirectoryProfiler(search_backend="rust", show_progress=False)
            rust_result = profiler_rust.probe(test_directory)
            results["rust"] = rust_result
        except Exception:
            pass

        # fd (if available)
        if fd.is_available():
            profiler_fd = DirectoryProfiler(search_backend="fd", show_progress=False)
            results["fd"] = profiler_fd.probe(test_directory)

        # Compare results between backends
        if len(results) >= 2:
            backend_names = list(results.keys())
            base_result = results[backend_names[0]]

            for backend_name in backend_names[1:]:
                compare_result = results[backend_name]

                # File counts should be very close (within small margin for hidden files)
                base_files = base_result["summary"]["total_files"]
                compare_files = compare_result["summary"]["total_files"]
                assert abs(base_files - compare_files) <= 2, f"File count mismatch: {backend_names[0]}={base_files}, {backend_name}={compare_files}"

                # Folder counts should be close
                base_folders = base_result["summary"]["total_folders"]
                compare_folders = compare_result["summary"]["total_folders"]
                assert abs(base_folders - compare_folders) <= 2, (
                    f"Folder count mismatch: {backend_names[0]}={base_folders}, {backend_name}={compare_folders}"
                )

    def test_performance_comparison(self, test_directory):
        """Compare performance between backends."""
        fd = FdIntegration()

        performance_results = {}

        # Test Python
        profiler_py = DirectoryProfiler(search_backend="python", show_progress=False)
        start_time = time.time()
        result_py = profiler_py.probe(test_directory)
        py_time = time.time() - start_time
        performance_results["python"] = {"time": py_time, "files": result_py["summary"]["total_files"]}

        # Test Rust (if available)
        try:
            profiler_rust = DirectoryProfiler(search_backend="rust", show_progress=False)
            start_time = time.time()
            result_rust = profiler_rust.probe(test_directory)
            rust_time = time.time() - start_time
            performance_results["rust"] = {"time": rust_time, "files": result_rust["summary"]["total_files"]}
        except Exception:
            pass

        # Test fd (if available)
        if fd.is_available():
            profiler_fd = DirectoryProfiler(search_backend="fd", show_progress=False)
            start_time = time.time()
            result_fd = profiler_fd.probe(test_directory)
            fd_time = time.time() - start_time
            performance_results["fd"] = {"time": fd_time, "files": result_fd["summary"]["total_files"]}

        # Verify we have results
        assert len(performance_results) >= 1

        # Print performance comparison for manual verification
        print("\n🚀 Performance Comparison:")
        for backend, perf in performance_results.items():
            files_per_sec = perf["files"] / perf["time"] if perf["time"] > 0 else 0
            print(f"  {backend}: {perf['time']:.3f}s ({perf['files']} files, {files_per_sec:.0f} files/sec)")

    def test_auto_backend_selection(self, test_directory):
        """Test that auto backend selection works correctly."""
        profiler = DirectoryProfiler(show_progress=False)  # Uses auto selection
        result = profiler.probe(test_directory)

        # Should successfully probe the directory
        assert result["summary"]["total_files"] >= 12
        assert result["summary"]["total_folders"] >= 8

        # Verify which backend was chosen
        chosen_backend = profiler._choose_backend()
        print(f"\n🤖 Auto selection chose: {chosen_backend}")

        # Should prefer Rust > fd > Python unless fd is explicitly enabled
        fd = FdIntegration()
        # Prefer Rust when available and enabled
        if hasattr(profiler, "_probe_rust") and profiler.use_rust:
            assert chosen_backend == "rust", "Auto selection should prefer Rust when available"
        # Otherwise prefer fd when enabled and available
        elif profiler.use_fd and fd.is_available():
            assert chosen_backend == "fd", "Auto selection should prefer fd when explicitly enabled"
        else:
            assert chosen_backend == "python", "Auto selection should fall back to Python"

    def test_max_depth_consistency(self, test_directory):
        """Test max_depth parameter across backends."""
        max_depth = 2

        # Test available backends with max_depth
        backends_to_test = []

        # Python
        profiler_py = DirectoryProfiler(search_backend="python", show_progress=False)
        backends_to_test.append(("python", profiler_py))

        # Rust (if available)
        try:
            profiler_rust = DirectoryProfiler(search_backend="rust", show_progress=False)
            backends_to_test.append(("rust", profiler_rust))
        except Exception:
            pass

        # fd (if available)
        fd = FdIntegration()
        if fd.is_available():
            profiler_fd = DirectoryProfiler(search_backend="fd", show_progress=False)
            backends_to_test.append(("fd", profiler_fd))

        results = {}
        for backend_name, profiler in backends_to_test:
            result = profiler.probe(test_directory, max_depth=max_depth)
            results[backend_name] = result

        # Compare results between backends
        if len(results) >= 2:
            backend_names = list(results.keys())

            for i in range(len(backend_names)):
                for j in range(i + 1, len(backend_names)):
                    backend1, backend2 = backend_names[i], backend_names[j]
                    result1, result2 = results[backend1], results[backend2]

                    # Results should be reasonably close
                    files_diff = abs(result1["summary"]["total_files"] - result2["summary"]["total_files"])
                    folders_diff = abs(result1["summary"]["total_folders"] - result2["summary"]["total_folders"])

                    # Debug output for failed tests
                    if files_diff > 6:
                        print(f"\nDEBUG: Large file count difference between {backend1} and {backend2}")
                        print(f"  {backend1}: {result1['summary']['total_files']} files")
                        print(f"  {backend2}: {result2['summary']['total_files']} files")
                        print(f"  Difference: {files_diff}")

                    assert files_diff <= 6, (
                        f"Max depth file count difference too large: {backend1} vs {backend2} "
                        f"(files: {result1['summary']['total_files']} vs {result2['summary']['total_files']})"
                    )
                    assert folders_diff <= 4, (
                        f"Max depth folder count difference too large: {backend1} vs {backend2} "
                        f"(folders: {result1['summary']['total_folders']} vs {result2['summary']['total_folders']})"
                    )


class TestFdFinder:
    """Test FdFinder interface thoroughly."""

    @pytest.fixture
    def test_directory(self):
        """Create test directory for FdFinder tests."""
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)

            # Create test files
            files = {
                "test.py": "print('hello')",
                "main.py": "def main(): pass",
                "config.json": '{"setting": true}',
                "data.csv": "col1,col2\n1,2",
                "README.md": "# Test",
                "nested/deep.py": "# nested python file",
                "nested/image.png": "fake png",
                ".hidden.txt": "hidden file",
            }

            for file_path, content in files.items():
                full_path = tmp_path / file_path
                full_path.parent.mkdir(parents=True, exist_ok=True)
                full_path.write_text(content)

            yield str(tmp_path)

    def test_fd_finder_basic(self, test_directory):
        """Test basic FdFinder functionality."""
        searcher = FdFinder()
        if not searcher.is_available():
            pytest.skip("fd not available")

        # Test finding Python files
        py_files = searcher.find_files(pattern=".*\\.py$", path=test_directory)
        assert len(py_files) >= 3  # test.py, main.py, nested/deep.py

        # Test finding by extension
        json_files = searcher.find_by_extension([".json"], path=test_directory)
        assert len(json_files) >= 1  # config.json

    def test_fd_finder_patterns(self, test_directory):
        """Test different pattern types in FdFinder."""
        searcher = FdFinder()
        if not searcher.is_available():
            pytest.skip("fd not available")

        # Test glob pattern
        py_files_glob = searcher.find_files(pattern="*.py", path=test_directory, use_glob=True)
        assert len(py_files_glob) >= 2

        # Test regex pattern
        py_files_regex = searcher.find_files(pattern=".*\\.py$", path=test_directory, use_glob=False)
        assert len(py_files_regex) >= 2

        # Test case insensitive
        readme_files = searcher.find_files(pattern="readme", path=test_directory, case_sensitive=False)
        assert len(readme_files) >= 1

    def test_fd_finder_directories(self, test_directory):
        """Test directory finding with FdFinder."""
        searcher = FdFinder()
        if not searcher.is_available():
            pytest.skip("fd not available")

        # Find directories
        dirs = searcher.find_directories(pattern="nested", path=test_directory)
        assert len(dirs) >= 1
        assert any("nested" in str(d) for d in dirs)

    def test_fd_finder_hidden_files(self, test_directory):
        """Test hidden file handling."""
        searcher = FdFinder()
        if not searcher.is_available():
            pytest.skip("fd not available")

        # Find hidden files
        hidden_files = searcher.find_files(pattern=".*", path=test_directory, hidden=True)

        # Should find at least our .hidden.txt file
        hidden_names = [Path(f).name for f in hidden_files]
        assert any(name.startswith(".") for name in hidden_names)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from abc import ABC, abstractmethod
from typing import Union

from rich.console import Console
from rich.table import Table

from .image_profiler import ImageReport


class BaseImageProfiler(ABC):
    def __init__(self):
        self.console = Console()

    @abstractmethod
    def probe(self, path):
        """Perform analysis and return an ImageReport."""
        pass

    def print_report(self, report: Union[dict, "ImageReport"]):
        """Print a formatted report of the analysis.

        Accepts either a dict or an `ImageReport` dataclass.
        """
        # Support both dicts and ImageReport instances
        if hasattr(report, "to_dict"):
            data = report.to_dict()
        else:
            data = dict(report)

        table = Table(title=f"Image Analysis Report: {data.get('path', 'Unknown')}")
        table.add_column("Property", style="cyan", no_wrap=True)
        table.add_column("Value", style="magenta")

        for key, value in data.items():
            if key != "path":  # Don't duplicate the path in the table
                table.add_row(str(key), str(value))

        self.console.print(table)
